import numpy as np
from qubx.utils import sbp


def percentile_rank(x: np.ndarray, v, pctls=np.arange(1, 101)):
    """
    Find percentile rank of value v
    :param x: values array
    :param v: vakue to be ranked
    :param pctls: percentiles
    :return: rank

    >>> percentile_rank(np.random.randn(1000), 1.69)
    >>> 95
    >>> percentile_rank(np.random.randn(1000), 1.69, [10,50,100])
    >>> 2
    """
    return np.argmax(np.sign(np.append(np.percentile(x, pctls), np.inf) - v))


def compare_to_norm(xs, xranges=None):
    """
    Compare distribution from xs against normal using estimated mean and std
    """
    import scipy.stats as stats
    import matplotlib.pyplot as plt
    import seaborn as sns

    _m, _s = np.mean(xs), np.std(xs)
    fit = stats.norm.pdf(sorted(xs), _m, _s)

    sbp(12, 1)
    plt.plot(sorted(xs), fit, 'r--', lw=2, label='N(%.2f, %.2f)' % (_m, _s))
    plt.legend(loc='upper right')

    sns.kdeplot(xs, color='g', label='Data', shade=True)
    if xranges is not None and len(xranges) > 1:
        plt.xlim(xranges)
    plt.legend(loc='upper right')

    sbp(12, 2)
    stats.probplot(xs, dist="norm", sparams=(_m, _s), plot=plt)



import glob, re
import json, os, dataclasses
from datetime import datetime
from typing import Dict, List, Optional
import configparser
import stackprinter

from qubx.core.basics import Instrument, FuturesInfo, TransactionCostsCalculator
from qubx.utils.marketdata.binance import get_binance_symbol_info_for_type
from qubx import logger
from qubx.utils.misc import makedirs, get_local_qubx_folder

_DEF_INSTRUMENTS_FOLDER = "instruments"
_DEF_FEES_FOLDER = "fees"


class _InstrumentEncoder(json.JSONEncoder):
    def default(self, obj):
        if dataclasses.is_dataclass(obj):
            return {k:v for k,v in dataclasses.asdict(obj).items() if not k.startswith('_')}
        if isinstance(obj, (datetime)):
            return obj.isoformat()
        return super().default(obj)


class _InstrumentDecoder(json.JSONDecoder):
    def _preprocess(d, ks):
        fi = d.get(ks)
        _preproc = lambda x: x[:x.find('.')] if x.endswith('Z') else x
        if fi:
            fi['delivery_date'] = datetime.strptime(_preproc(fi.get('delivery_date', '5000-01-01T00:00:00')),'%Y-%m-%dT%H:%M:%S')
            fi['onboard_date'] = datetime.strptime(_preproc(fi.get('onboard_date', '1970-01-01T00:00:00')),'%Y-%m-%dT%H:%M:%S')
        return d | {ks: FuturesInfo(**fi) if fi else None}

    def decode(self, json_string):
        obj = super(_InstrumentDecoder, self).decode(json_string)
        if isinstance(obj, dict):
            return Instrument(**_InstrumentDecoder._preprocess(obj, 'futures_info'))
        elif isinstance(obj, list):
            return [Instrument(**_InstrumentDecoder._preprocess(o, 'futures_info')) for o in obj]
        return obj


class InstrumentsLookup:
    _lookup: Dict[str, Instrument]
    _path: str

    def __init__(self, path: str=makedirs(get_local_qubx_folder(), _DEF_INSTRUMENTS_FOLDER)) -> None:
        self._path = path
        if not self.load():
            self.refresh()
        self.load()

    def load(self) -> bool:
        self._lookup = {}
        data_exists = False
        for fs in glob.glob(self._path + '/*.json'):
            try:
                with open(fs, 'r') as f:
                    instrs = json.load(f, cls=_InstrumentDecoder)
                    for i in instrs:  
                        self._lookup[f"{i.exchange}:{i.symbol}"] = i
                    data_exists = True
            except Exception as ex:
                stackprinter.show_current_exception()
                logger.warning(ex)

        return data_exists

    def find(self, exchange: str, base: str, quote: str, margin: Optional[str]=None) -> Optional[Instrument]:
        for i in self._lookup.values():
            if i.exchange == exchange and (
                (i.base == base and i.quote == quote) or (i.base == quote and i.quote == base)
            ):
                if margin is not None and i.margin_symbol is not None:
                    if i.margin_symbol == margin:
                        return i
                else:
                    return i
        return None

    def find_symbol(self, exchange: str, symbol: str) -> Optional[Instrument]:
        for i in self._lookup.values():
            if (i.exchange == exchange) and (i.symbol == symbol):
                return i
        return None
    
    def _save_to_json(self, path, instruments: List[Instrument]):
        with open(path, 'w') as f:
            json.dump(instruments, f, cls=_InstrumentEncoder)
        logger.info(f'Saved {len(instruments)} to {path}')

    def find_aux_instrument_for(self, instrument: Instrument, base_currency: str) -> Optional[Instrument]:
        """
        Tries to find aux instrument (for conversions to funded currency)
        for example: 
            ETHBTC -> BTCUSDT for base_currency USDT
            EURGBP -> GBPUSD for base_currency USD
            ...
        """
        base_currency = base_currency.upper()
        if instrument.quote != base_currency and instrument._aux_instrument is None:
            return self.find(instrument.exchange, instrument.quote, base_currency)
        return instrument._aux_instrument

    def __getitem__(self, spath: str) -> List[Instrument]:
        res = []
        c = re.compile(spath)
        for k, v in self._lookup.items():
            if re.match(c, k):
                res.append(v)
        return res
    
    def refresh(self):
        for mn in dir(self):
            if mn.startswith('_update_'):
                getattr(self, mn)(self._path)

    def _update_kraken(self, path: str):
        import ccxt as cx
        kf = cx.krakenfutures()
        ks = cx.kraken()
        f_mkts = kf.load_markets()
        s_mkts = ks.load_markets()

        # - process futures
        f_instruments = []
        for _, v in f_mkts.items():
            info = v['info']
            # - we skip index as it's not traded
            if v['index']: continue
            maint_margin=0
            required_margin=0
            if 'marginLevels' in info:
                margins = info['marginLevels'][0]
                maint_margin=float(margins['maintenanceMargin'])
                required_margin=float(margins['initialMargin'])
            f_instruments.append(
                Instrument(
                    v['symbol'], v['type'], 'KRAKEN.F', v['base'], v['quote'], v['settle'],
                    min_tick=float(info['tickSize']),
                    min_size_step=float(v['precision']['price']),
                    min_size=v['precision']['amount'],
                    futures_info=FuturesInfo(
                        contract_type=info['type'],
                        contract_size=float(info['contractSize']),
                        onboard_date=info['openingDate'],
                        delivery_date=v['expiryDatetime'] if 'expiryDatetime' in info else "2100-01-01T00:00:00",
                        maint_margin=maint_margin, required_margin=required_margin,
                    )
                )
            )
        # - drop to file
        self._save_to_json(os.path.join(path, 'kraken.f.json'), f_instruments)

        # - process spots
        s_instruments = []
        for _, v in s_mkts.items():
            info = v['info']
            s_instruments.append(
                Instrument(
                    v['symbol'], v['type'], 'KRAKEN', v['base'], v['quote'], v['settle'],
                    min_tick=float(info['tick_size']), 
                    min_size_step=float(v['precision']['price']), 
                    min_size=float(v['precision']['amount']),
                )
            )
        # - drop to file
        self._save_to_json(os.path.join(path, 'kraken.json'), s_instruments)

    def _update_dukas(self, path: str):
        instruments = [
            Instrument('EURUSD', 'FX', 'DUKAS', 'EUR', 'USD', 'USD', 0.00001, 1, 1000),
            Instrument('GBPUSD', 'FX', 'DUKAS', 'GBP', 'USD', 'USD', 0.00001, 1, 1000),
            Instrument('USDJPY', 'FX', 'DUKAS', 'USD', 'JPY', 'USD', 0.001,   1, 1000),
            Instrument('USDCAD', 'FX', 'DUKAS', 'USD', 'CAD', 'USD', 0.00001, 1, 1000),
            Instrument('AUDUSD', 'FX', 'DUKAS', 'AUD', 'USD', 'USD', 0.00001, 1, 1000),
            Instrument('USDPLN', 'FX', 'DUKAS', 'USD', 'PLN', 'USD', 0.00001, 1, 1000),
            Instrument('EURGBP', 'FX', 'DUKAS', 'EUR', 'GBP', 'USD', 0.00001, 1, 1000),
            # TODO: addd all or find how to get it from site
        ]
        self._save_to_json(os.path.join(path, 'dukas.json'), instruments)

    def _update_binance(self, path: str):
        infos = get_binance_symbol_info_for_type(['UM', 'CM', 'SPOT'])
        for exchange, info in infos.items():
            instruments = []
            for s in info['symbols']:
                tick_size, size_step = None, None 
                for i in s['filters']:
                    if i['filterType'] == 'PRICE_FILTER':
                        tick_size = float(i['tickSize'])
                    if i['filterType'] == 'LOT_SIZE':
                        size_step = float(i['stepSize'])

                fut_info = None
                if 'contractType' in s:
                    fut_info = FuturesInfo( 
                        s.get('contractType', 'UNKNOWN'),
                        datetime.fromtimestamp(s.get('deliveryDate', 0)/1000.0),
                        datetime.fromtimestamp(s.get('onboardDate', 0)/1000.0),
                        float(s.get('contractSize', 1)), 
                        float(s.get('maintMarginPercent', 0)), 
                        float(s.get('requiredMarginPercent', 0)), 
                        float(s.get('liquidationFee', 0)),
                    )

                instruments.append(
                    Instrument(
                        s['symbol'], 'CRYPTO', exchange.upper(), s['baseAsset'], s['quoteAsset'], s.get('marginAsset', None), 
                        tick_size, size_step, 
                        min_size=size_step,    # TODO: not sure about minimal position for Binance
                        futures_info=fut_info
                ))
            # - store to file
            self._save_to_json(os.path.join(path, f'{exchange}.json'), instruments)

# - TODO: need to find better way to extract actual data !!
_DEFAULT_FEES = """
[binance]
# SPOT (maker, taker)
vip0_usdt = 0.1000,0.1000
vip1_usdt = 0.0900,0.1000
vip2_usdt = 0.0800,0.1000
vip3_usdt = 0.0420,0.0600
vip4_usdt = 0.0420,0.0540
vip5_usdt = 0.0360,0.0480
vip6_usdt = 0.0300,0.0420
vip7_usdt = 0.0240,0.0360
vip8_usdt = 0.0180,0.0300
vip9_usdt = 0.0120,0.0240

# SPOT (maker, taker)
vip0_bnb = 0.0750,0.0750
vip1_bnb = 0.0675,0.0750
vip2_bnb = 0.0600,0.0750
vip3_bnb = 0.0315,0.0450
vip4_bnb = 0.0315,0.0405
vip5_bnb = 0.0270,0.0360
vip6_bnb = 0.0225,0.0315
vip7_bnb = 0.0180,0.0270
vip8_bnb = 0.0135,0.0225
vip9_bnb = 0.0090,0.0180

# UM futures (maker, taker)
[binance.um]
vip0_usdt = 0.0200,0.0500
vip1_usdt = 0.0160,0.0400
vip2_usdt = 0.0140,0.0350
vip3_usdt = 0.0120,0.0320
vip4_usdt = 0.0100,0.0300
vip5_usdt = 0.0080,0.0270
vip6_usdt = 0.0060,0.0250
vip7_usdt = 0.0040,0.0220
vip8_usdt = 0.0020,0.0200
vip9_usdt = 0.0000,0.0170

# CM futures (maker, taker)
[binance.cm]
vip0 = 0.0200,0.0500
vip1 = 0.0160,0.0400
vip2 = 0.0140,0.0350
vip3 = 0.0120,0.0320
vip4 = 0.0100,0.0300
vip5 = 0.0080,0.0270
vip6 = 0.0060,0.0250
vip7 = 0.0040,0.0220
vip8 = 0.0020,0.0200
vip9 = 0.0000,0.0170

[bitmex]
tierb_xbt=0.02,0.075
tierb_usdt=-0.015,0.075
tieri_xbt=0.01,0.05
tieri_usdt=-0.015,0.05
tiert_xbt=0.0,0.04
tiert_usdt=-0.015,0.04
tierm_xbt=0.0,0.035
tierm_usdt=-0.015,0.035
tiere_xbt=0.0,0.03
tiere_usdt=-0.015,0.03
tierx_xbt=0.0,0.025
tierx_usdt=-0.015,0.025
tierd_xbt=-0.003,0.024
tierd_usdt=-0.015,0.024
tierw_xbt=-0.005,0.023
tierw_usdt=-0.015,0.023
tierk_xbt=-0.008,0.022
tierk_usdt=-0.015,0.022
tiers_xbt=-0.01,0.0175
tiers_usdt=-0.015,0.02

[dukas]
regular=0.0035,0.0035
premium=0.0017,0.0017

[kraken]
K0=0.25,0.40
K10=0.20,0.35
K50=0.14,0.24
K100=0.12,0.22
K250=0.10,0.20
K500=0.08,0.18
M1=0.06,0.16
M2.5=0.04,0.14
M5=0.02,0.12
M10=0.0,0.10

[kraken.f]
K0=0.0200,0.0500
K100=0.0150,0.0400
M1=0.0125,0.0300
M5=0.0100,0.0250
M10=0.0075,0.0200
M20=0.0050,0.0150
M50=0.0025,0.0125
M100=0.0000,0.0100
"""


class FeesLookup:
    """
    Fees lookup
    """
    _lookup: Dict[str, TransactionCostsCalculator]
    _path: str

    def __init__(self, path: str=makedirs(get_local_qubx_folder(), _DEF_FEES_FOLDER)) -> None:
        self._path = path
        if not self.load():
            self.refresh()
        self.load()

    def load(self) -> bool:
        self._lookup = {}
        data_exists = False
        parser = configparser.ConfigParser()
        # - load all avaliable configs
        for fs in glob.glob(self._path + '/*.ini'):
            parser.read(fs)
            data_exists = True

        for exch in parser.sections():
            for spec, info in parser[exch].items():
                try:
                    maker, taker = info.split(',')
                    self._lookup[f"{exch}_{spec}"] = (float(maker), float(taker))
                except:
                    logger.warning(f'Wrong spec format for {exch}: "{info}". Should be spec=maker,taker')

        return data_exists

    def __getitem__(self, spath: str) -> List[Instrument]:
        res = []
        c = re.compile(spath)
        for k, v in self._lookup.items():
            if re.match(c, k):
                res.append((k,v))
        return res
    
    def refresh(self):
        with open(os.path.join(self._path, 'default.ini'), 'w') as f:
            f.write(_DEFAULT_FEES)

    def find(self, exchange: str, spec: str) -> Optional[TransactionCostsCalculator]:
        key = f"{exchange}_{spec}"
        vals = self._lookup.get(key)
        return TransactionCostsCalculator(key, *self._lookup.get(key)) if vals is not None else None

    def __repr__(self) -> str:
        s = "Name:\t\t\t(maker, taker)\n"
        for k, v in self._lookup.items():
            s += f"{k.ljust(25)}: {v}\n"
        return s


@dataclasses.dataclass(frozen=True)
class GlobalLookup:
    instruments: InstrumentsLookup
    fees: FeesLookup

    def find_fees(self, exchange: str, spec: str) -> Optional[TransactionCostsCalculator]:
        return self.fees.find(exchange, spec)

    def find_aux_instrument_for(self, instrument: Instrument, base_currency: str) -> Optional[Instrument]:
        return self.instruments.find_aux_instrument_for(instrument, base_currency)

    def find_instrument(self, exchange: str, base: str, quote: str) -> Optional[Instrument]:
        return self.instruments.find(exchange, base, quote)

    def find_symbol(self, exchange: str, symbol: str) -> Optional[Instrument]:
        return self.instruments.find_symbol(exchange, symbol)


""""
Here stuff we want to have in every Jupyter notebook after calling %qube magic
"""
import importlib_metadata

import qubx
from qubx.utils import runtime_env
from qubx.utils.misc import add_project_to_system_path, logo


def np_fmt_short():
    # default np output is 75 columns so extend it a bit and suppress scientific fmt for small floats
    np.set_printoptions(linewidth=240, suppress=True)


def np_fmt_reset():
    # reset default np printing options
    np.set_printoptions(edgeitems=3, infstr='inf', linewidth=75, nanstr='nan', precision=8,
                        suppress=False, threshold=1000, formatter=None)


if runtime_env() in ['notebook', 'shell']:

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # -- all imports below will appear in notebook after calling %%alphalab magic ---
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    # - - - - Common stuff - - - -
    import numpy as np
    import pandas as pd
    from datetime import time, timedelta
    from tqdm.auto import tqdm

    # - - - - TA stuff and indicators - - - -
    # - - - - Portfolio analysis - - - -
    # - - - - Simulator stuff - - - -
    # - - - - Learn stuff - - - -
    # - - - - Charting stuff - - - -
    from matplotlib import pyplot as plt
    from qubx.utils.charting.mpl_helpers import fig, subplot, sbp
    # - - - - Utils - - - -

    # - setup short numpy output format
    np_fmt_short()
    
    # - add project home to system path
    add_project_to_system_path()

    # show logo first time
    if not hasattr(qubx.QubxMagics, '__already_initialized__'):
        setattr(qubx.QubxMagics, "__already_initialized__", True)
        logo()




from threading import Thread
from typing import Any, Dict, List, Optional

import asyncio
from asyncio.tasks import Task
from asyncio.events import AbstractEventLoop
from collections import defaultdict
import stackprinter

import ccxt.pro as cxp
from ccxt.base.exchange import Exchange
from ccxt import NetworkError, ExchangeClosedByUser

import re
import pandas as pd

from qubx import logger
from qubx.core.basics import Instrument, Position, dt_64, Deal
from qubx.core.strategy import IDataProvider, CtrlChannel, IExchangeServiceProvider
from qubx.core.series import TimeSeries, Bar, Trade, Quote
from qubx.impl.utils import DATA_PROVIDERS_ALIASES

from .ccxt_trading import CCXTSyncTradingConnector

# - register custom wrappers
from .exchange_customizations import BinanceQV
cxp.binanceqv = BinanceQV            # type: ignore
cxp.exchanges.append('binanceqv')


class CCXTConnector(IDataProvider, CCXTSyncTradingConnector):
    _exchange: Exchange
    _subsriptions: Dict[str, List[str]]

    _ch_market_data: CtrlChannel
    _last_quotes: Dict[str, Optional[Quote]]
    _loop: AbstractEventLoop 
    _thread_event_loop: Thread

    def __init__(self, exchange_id: str, account_id: str, base_currency: str | None = None, commissions: str|None = None, **exchange_auth):
        super().__init__(exchange_id, account_id, base_currency, commissions, **exchange_auth)
        
        exchange_id = exchange_id.lower()
        exch = DATA_PROVIDERS_ALIASES.get(exchange_id, exchange_id)
        if exch not in cxp.exchanges:
            raise ValueError(f"Exchange {exchange_id} -> {exch} is not supported by CCXT.pro !")

        # - create new even loop
        self._loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self._loop)

        # - create exchange's instance
        self._exchange = getattr(cxp, exch)(exchange_auth | {'asyncio_loop': self._loop})
        self._ch_market_data = CtrlChannel(exch + '.marketdata', sent=(None, None, None))
        self._last_quotes = defaultdict(lambda: None)
        self._subsriptions: Dict[str, List[str]] = defaultdict(list)

        logger.info(f"{self.get_name().upper()} initialized - current time {self.time()}")

    def subscribe(self, subscription_type: str, symbols: List[str], timeframe:Optional[str]=None, nback:int=0) -> bool:
        to_process = self._check_existing_subscription(subscription_type.lower(), symbols)
        if not to_process:
            logger.info(f"Symbols {symbols} already subscribed on {subscription_type} data")
            return False

        # - subscribe to market data updates
        match sbscr := subscription_type.lower():
            case 'ohlc':
                if timeframe is None:
                    raise ValueError("timeframe must not be None for OHLC data subscription")

                # convert to exchange format
                tframe = self._get_exch_timeframe(timeframe)
                for s in to_process:
                    # self._task_a(self._listen_to_ohlcv(self.get_communication_channel(), s, tframe, nback))
                    asyncio.run_coroutine_threadsafe(self._listen_to_ohlcv(self.get_communication_channel(), s, tframe, nback), self._loop)
                    self._subsriptions[sbscr].append(s.lower())
                logger.info(f'Subscribed on {sbscr} updates for {len(to_process)} symbols: \n\t\t{to_process}')

            case 'trades':
                raise ValueError("TODO")

            case 'quotes':
                raise ValueError("TODO")

            case _:
                raise ValueError("TODO")

        # - subscibe to executions reports
        for s in to_process:
            asyncio.run_coroutine_threadsafe(self._listen_to_execution_reports(self.get_communication_channel(), s), self._loop)

        self._thread_event_loop = Thread(target=self._loop.run_forever, args=(), daemon=True)
        self._thread_event_loop.start()

        return True

    def get_communication_channel(self) -> CtrlChannel:
        return self._ch_market_data

    def _check_existing_subscription(self, subscription_type, symbols: List[str]) -> List[str]:
        subscribed = self._subsriptions[subscription_type]
        to_subscribe = []
        for s in symbols: 
            if s not in subscribed:
                to_subscribe.append(s)
        return to_subscribe

    def _time_msec_nbars_back(self, timeframe: str, nbarsback: int) -> int:
        return (self.time() - nbarsback * pd.Timedelta(timeframe)).asm8.item() // 1000000

    def get_historical_ohlcs(self, symbol: str, timeframe: str, nbarsback: int) -> Optional[List[Bar]]:
        assert nbarsback >= 1
        since = self._time_msec_nbars_back(timeframe, nbarsback)

        # - get this from sync 
        res = self.sync.fetch_ohlcv(symbol, self._get_exch_timeframe(timeframe), since=since)
        _arr = []  
        for oh in res: # type: ignore
            _arr.append(
                Bar(oh[0] * 1_000_000, oh[1], oh[2], oh[3], oh[4], oh[6], oh[7]) 
                if len(oh) > 6 else 
                Bar(oh[0] * 1_000_000, oh[1], oh[2], oh[3], oh[4], oh[5]) 
            )
        return _arr

    async def _listen_to_execution_reports(self, channel: CtrlChannel, symbol: str):
        while channel.control.is_set():
            try:
                exec = await self._exchange.watch_orders(symbol)        # type: ignore
                _msg = f"\nexecs_{symbol} = [\n"
                for report in exec:
                    _msg += '\t' + str(report) + ',\n'
                    order, deals = self._process_execution_report(symbol, report)
                    # - send update to client 
                    channel.queue.put((symbol, 'order', order))
                    if deals: 
                        channel.queue.put((symbol, 'deals', deals))
                logger.debug(_msg + "]\n")

            except NetworkError as e:
                logger.error(f"(CCXTConnector) NetworkError in _listen_to_execution_reports : {e}")
                await asyncio.sleep(1)
                continue

            except ExchangeClosedByUser:
                # - we closed connection so just stop it 
                logger.info(f"(CCXTConnector) {symbol} execution reports listening has been stopped")
                break

            except Exception as err:
                logger.error(f"(CCXTConnector) exception in _listen_to_execution_reports : {err}")
                logger.error(stackprinter.format(err))

    async def _listen_to_ohlcv(self, channel: CtrlChannel, symbol: str, timeframe: str, nbarsback: int):
        # - check if we need to load initial 'snapshot'
        if nbarsback >= 1:
            start = self._time_msec_nbars_back(timeframe, nbarsback)
            ohlcv = await self._exchange.fetch_ohlcv(symbol, timeframe, since=start, limit=nbarsback + 1)        # type: ignore
            # - just send data as the list
            channel.queue.put((
                symbol, 
                'hist_bars', 
                [Bar(oh[0] * 1_000_000, oh[1], oh[2], oh[3], oh[4], oh[6], oh[7]) for oh in ohlcv]
            ))
            logger.info(f"{symbol}: loaded {len(ohlcv)} {timeframe} bars")

        while channel.control.is_set():
            try:
                ohlcv = await self._exchange.watch_ohlcv(symbol, timeframe)        # type: ignore

                # - update positions by actual close price
                self.update_position_price(symbol, ohlcv[-1][4])

                for oh in ohlcv:
                    channel.queue.put((symbol, 'bar', Bar(oh[0] * 1000000, oh[1], oh[2], oh[3], oh[4], oh[6], oh[7])))

            except NetworkError as e:
                logger.error(f"(CCXTConnector) NetworkError in _listen_to_ohlcv : {e}")
                await asyncio.sleep(1)
                continue

            except ExchangeClosedByUser:
                # - we closed connection so just stop it 
                logger.info(f"(CCXTConnector) {symbol} OHLCV listening has been stopped")
                break

            except Exception as e:
                # logger.error(str(e))
                logger.error(f"(CCXTConnector) exception in _listen_to_ohlcv : {e}")
                logger.error(stackprinter.format(e))
                await self._exchange.close()        # type: ignore
                raise e

    def get_quote(self, symbol: str) -> Optional[Quote]:
        return self._last_quotes[symbol]

    def _get_exch_timeframe(self, timeframe: str):
        if timeframe is not None:
            _t = re.match(r'(\d+)(\w+)', timeframe)
            timeframe = f"{_t[1]}{_t[2][0].lower()}" if _t and len(_t.groups()) > 1 else timeframe

        tframe = self._exchange.find_timeframe(timeframe)
        if tframe is None:
            raise ValueError(f"timeframe {timeframe} is not supported by {self.get_name()}")

        return tframe

    def close(self):
        try:
            self._loop.run_until_complete(self._exchange.close()) # type: ignore
        except Exception as e:
            logger.error(e)



"""
 # All interfaces related to strategy etc
"""
from collections import defaultdict
from threading import Thread
import traceback
from typing import Any, Callable, Dict, List, Optional, Tuple, Union
from types import FunctionType
import numpy as np
from dataclasses import dataclass
from enum import Enum

import pandas as pd

from qubx import lookup, logger
from qubx.core.account import AccountProcessor
from qubx.core.loggers import BalanceLogger, ExecutionsLogger, LogsWriter, PortfolioLogger, PositionsDumper, StrategyLogging
from qubx.core.lookups import InstrumentsLookup
from qubx.core.basics import Deal, Instrument, Order, Position, Signal, dt_64, td_64, CtrlChannel
from qubx.core.series import TimeSeries, Trade, Quote, Bar, OHLCV
from qubx.utils.time import convert_tf_str_td64
from qubx.utils.misc import Stopwatch


@dataclass
class EventFromExchange:
    time: dt_64
    type: str
    data: Optional[Any] 


@dataclass
class TriggerEvent:
    time: dt_64
    type: str
    instrument: Optional[Instrument]
    data: Optional[Any] 


class IDataProvider:

    def subscribe(self, subscription_type: str, symbols: List[str], **kwargs) -> bool:
        raise NotImplementedError("subscribe")

    def get_communication_channel(self) -> CtrlChannel:
        raise NotImplementedError("get_communication_channel")

    def get_historical_ohlcs(self, symbol: str, timeframe: str, nbarsback: int) -> List[Bar] | None:
        raise NotImplementedError("get_historical_ohlcs")

    def get_quote(self, symbol: str) -> Quote | None:
        raise NotImplementedError("get_quote")


class IExchangeServiceProvider:
    acc: AccountProcessor

    def time(self) -> dt_64:
        """
        Returns current time
        """
        raise NotImplementedError("time is not implemented")

    def get_account(self) -> AccountProcessor:
        return self.acc

    def get_name(self) -> str:
        raise NotImplementedError("get_name is not implemented")

    def schedule_trigger(self, trigger_id: str, when: str):
        raise NotImplementedError("schedule_trigger is not implemented")

    def get_capital(self) -> float:
        return self.acc.get_capital()

    def send_order(self, instrument: Instrument, order_side: str, order_type: str, amount: float, price: float | None = None, 
        client_id: str | None = None, time_in_force: str='gtc') -> Order | None:
        raise NotImplementedError("send_order is not implemented")

    def cancel_order(self, order_id: str) -> Order | None:
        raise NotImplementedError("cancel_order is not implemented")

    def get_orders(self, symbol: str | None = None) -> List[Order]:
        raise NotImplementedError("get_orders is not implemented")

    def get_position(self, instrument: Instrument) -> Position:
        raise NotImplementedError("get_position is not implemented")

    def get_base_currency(self) -> str:
        raise NotImplementedError("get_basic_currency is not implemented")


class PositionsTracker:
    ctx: 'StrategyContext'

    def __init__(self, ctx: 'StrategyContext') -> None:
        self.ctx = ctx
    

class IStrategy:
    ctx: 'StrategyContext'

    def on_start(self, ctx: 'StrategyContext'):
        pass

    def on_event(self, ctx: 'StrategyContext', event: TriggerEvent) -> Optional[List[Signal]]:
        return None

    def on_stop(self, ctx: 'StrategyContext'):
        pass

    def tracker(self, ctx: 'StrategyContext') -> PositionsTracker | None:
        pass


def _dict_with_exc(dct, f):
    if f not in dct:
        raise ValueError(f"Configuration {dct} must contain field '{f}'")
    return dct[f]


def _round_down_at_min_qty(x: float, min_size: float) -> float:
    return (int(x / min_size)) * min_size


_SW = Stopwatch()


class CachedMarketDataHolder: 
    _min_timeframe: dt_64
    _last_bar: Dict[str, Optional[Bar]]
    _ohlcvs: Dict[str, Dict[td_64, OHLCV]]

    def __init__(self, minimal_timeframe: str) -> None:
        self._min_timeframe = convert_tf_str_td64(minimal_timeframe)
        self._ohlcvs = dict()
        self._last_bar = defaultdict(lambda: None)

    def init_ohlcv(self, symbol: str, max_size=np.inf):
        self._ohlcvs[symbol] = {self._min_timeframe: OHLCV(symbol, self._min_timeframe, max_size)}
    
    @_SW.watch('CachedMarketDataHolder')
    def get_ohlcv(self, symbol: str, timeframe: str, max_size=np.inf) -> OHLCV:
        tf = convert_tf_str_td64(timeframe) 

        if symbol not in self._ohlcvs:
           self._ohlcvs[symbol] = {}

        if tf not in self._ohlcvs[symbol]: 
            # - check requested timeframe
            new_ohlc = OHLCV(symbol, tf, max_size)
            if tf < self._min_timeframe:
                logger.warning(f"[{symbol}] Request for timeframe {timeframe} that is smaller then minimal {self._min_timeframe}")
            else:
                # - first try to resample from smaller frame
                if (basis := self._ohlcvs[symbol].get(self._min_timeframe)):
                    for b in basis[::-1]:
                        new_ohlc.update_by_bar(b.time, b.open, b.high, b.low, b.close, b.volume, b.bought_volume)
                
            self._ohlcvs[symbol][tf] = new_ohlc

        return self._ohlcvs[symbol][tf]

    @_SW.watch('CachedMarketDataHolder')
    def update_by_bar(self, symbol: str, bar: Bar):
        _last_bar = self._last_bar[symbol]
        v_tot_inc = bar.volume
        v_buy_inc = bar.bought_volume

        if _last_bar is not None:
            if _last_bar.time == bar.time: # just current bar updated
                v_tot_inc -= _last_bar.volume
                v_buy_inc -= _last_bar.bought_volume

            if _last_bar.time > bar.time: # update is too late - skip it
                return

        if symbol in self._ohlcvs:
            self._last_bar[symbol] = bar
            for ser in self._ohlcvs[symbol].values():
                ser.update_by_bar(bar.time, bar.open, bar.high, bar.low, bar.close, v_tot_inc, v_buy_inc)

    @_SW.watch('CachedMarketDataHolder')
    def update_by_quote(self, symbol: str, quote: Quote):
        series = self._ohlcvs.get(symbol)
        if series:
            for ser in series.values():
                ser.update(quote.time, quote.mid_price(), 0)

    @_SW.watch('CachedMarketDataHolder')
    def update_by_trade(self, symbol: str, trade: Trade):
        series = self._ohlcvs.get(symbol)
        if series:
            total_vol = trade.size 
            bought_vol = total_vol if trade.taker >= 1 else 0.0 
            for ser in series.values():
                ser.update(trade.time, trade.price, total_vol, bought_vol)

 
class StrategyContext:
    MAX_NUMBER_OF_STRATEGY_FAILURES = 10

    strategy: IStrategy                         # strategy instance
    exchange_service: IExchangeServiceProvider  # service for exchange API: orders managemewnt
    data_provider: IDataProvider                # market data provider 
    instruments: List[Instrument]               # list of instruments this strategy trades
    positions: Dict[str, Position]              # positions of the strategy (instrument -> position)

    # - loggers
    _logging: StrategyLogging                    # recording all activities for the strat: execs, positions, portfolio

    # - cached marked data
    _cache: CachedMarketDataHolder # market data cache

    # - configuration
    _market_data_subcription_type:str = 'unknown'
    _market_data_subcription_params: dict = dict()
    _thread_data_loop: Thread | None = None            # market data loop

    _trig_interval_in_bar_nsec: int
    _trig_bar_freq_nsec: int
    _trig_on_bar: bool = False
    _trig_on_time: bool = False
    _trig_on_quote: bool = False
    _trig_on_trade: bool = False
    _trig_on_book: bool = False
    _current_bar_trigger_processed: bool = False
    _is_initilized: bool = False
    _symb_to_instr: Dict[str, Instrument]
    __strategy_id: str
    __order_id: int 
    __handlers: Dict[str, Callable[['StrategyContext', str, Any], TriggerEvent | None]]

    def __init__(self, 
            # - strategy with parameters
            strategy: IStrategy, config: Dict[str, Any] | None,
            # - - - - - - - - - - - - - - - - - - - - -

            # - data provider and exchange service
            data_provider: IDataProvider,
            exchange_service: IExchangeServiceProvider, 
            instruments: List[Instrument],
            # - - - - - - - - - - - - - - - - - - - - -

            # - context's parameters - - - - - - - - - -
            trigger: Dict[str, Any] = dict(type='ohlc', timeframe='1Min', nback=60),
            md_subscription: Dict[str,Any] = dict(type='ohlc', timeframe='1Min'),
            # - - - - - - - - - - - - - - - - - - - - -

            # - how to write logs - - - - - - - - - -
            logs_writer: LogsWriter | None = None,
            positions_log_freq: str = '1Min',
            portfolio_log_freq: str = '5Min',
            num_exec_records_to_write = 1      # in live let's write every execution 
        ) -> None:

        # - initialization
        self.exchange_service = exchange_service
        self.data_provider = data_provider
        self.config = config
        self.instruments = instruments
        self.positions = {}

        # - create strategy instance and populate custom paramameters
        self.__instantiate_strategy(strategy, config)

        # - for fast access to instrument by it's symbol
        self._symb_to_instr = {i.symbol: i for i in instruments}
 
        # - process trigger configuration
        self.__check_how_to_trigger_strategy(trigger)

        # - process market data configuration
        self.__check_how_to_listen_to_market_data(md_subscription)

        # - instantiate logging functional
        self._logging = StrategyLogging(
            logs_writer, positions_log_freq, portfolio_log_freq, num_exec_records_to_write
        )

        # - extract data and event handlers
        self.__handlers = {
            n.split('_processing_')[1]: f for n, f in self.__class__.__dict__.items() 
            if type(f) == FunctionType and n.startswith('_processing_') 
        }

    def __instantiate_strategy(self, strategy: IStrategy, config: Dict[str, Any] | None):
        # - set parameters to strategy (not sure we will do it here)
        self.strategy = strategy
        if isinstance(strategy, type):
            self.strategy = strategy()
        self.strategy.ctx = self

        # TODO: - trackers - - - - - - - - - - - - -
        # - here we need to instantiate trackers 
        # - need to think how to do it properly !!!
        # TODO: - trackers - - - - - - - - - - - - -

        # - set strategy custom parameters 
        self.populate_parameters_to_strategy(self.strategy, **config if config else {})
        self._is_initilized = False
        self.__strategy_id = self.strategy.__class__.__name__ + "_"
        self.__order_id = self.time().item() // 100_000_000

    def __check_how_to_listen_to_market_data(self, md_config: dict):
        self._market_data_subcription_type = _dict_with_exc(md_config, 'type').lower()
        match self._market_data_subcription_type:
            case 'ohlc':
                timeframe = _dict_with_exc(md_config, 'timeframe')
                self._market_data_subcription_params = {
                    'timeframe': timeframe,
                    'nback': md_config.get('nback', 1), 
                }
                self._cache = CachedMarketDataHolder(timeframe) 

            case 'trade' | 'trades' | 'tas':
                self._cache = CachedMarketDataHolder('1Sec') 

            case 'quote' | 'quotes':
                self._cache = CachedMarketDataHolder('1Sec') 

            case 'ob' | 'orderbook':
                self._cache = CachedMarketDataHolder('1Sec') 

            case _:
                raise ValueError(f"{self._market_data_subcription_type} is not a valid value for market data subcription type !!!")

    def __check_how_to_trigger_strategy(self, trigger_config: dict):
        # - check how it's configured to be triggered
        self._trig_interval_in_bar_nsec = 0

        match (_trigger := _dict_with_exc(trigger_config, 'type').lower()):
            case 'bar': 
                self._trig_on_bar = True

                _bar_timeframe = pd.Timedelta(_dict_with_exc(trigger_config, 'timeframe'))
                _inside_bar_delay = pd.Timedelta(_dict_with_exc(trigger_config, 'delay'))

                if abs(pd.Timedelta(_inside_bar_delay)) > pd.Timedelta(_bar_timeframe):
                    raise ValueError(f"Delay must be less or equal to bar's timeframe for {_trigger} trigger: you set delay {_inside_bar_delay} for {_bar_timeframe}")

                # for positive delay - trigger strategy when this interval passed after new bar's open
                if _inside_bar_delay >= pd.Timedelta(0): 
                    self._trig_interval_in_bar_nsec = _inside_bar_delay.asm8.item()
                # for negative delay - trigger strategy when time is closer to bar's closing time more than this interval
                else:
                    self._trig_interval_in_bar_nsec = (_bar_timeframe + _inside_bar_delay).asm8.item()
                self._trig_bar_freq_nsec = _bar_timeframe.asm8.item()

            case 'time': 
                self._trig_on_time = True
                _time_to_trigger = _dict_with_exc(trigger_config, 'when')

                # - schedule periodic timer
                self.exchange_service.schedule_trigger('time', _time_to_trigger)

            case 'quote': 
                self._trig_on_quote = True
                raise ValueError(f"{_trigger} NOT IMPLEMENTED")

            case 'trade': 
                self._trig_on_trade = True
                raise ValueError(f"{_trigger} NOT IMPLEMENTED")

            case 'orderbook' | 'ob': 
                self._trig_on_book = True
                raise ValueError(f"{_trigger} NOT IMPLEMENTED")

            case _: 
                raise ValueError(f"Wrong trigger type {_trigger}")
   
    def _process_incoming_data(self, channel: CtrlChannel):
        _fails_counter = 0 
        logger.info("(StrategyContext) Start processing market data")

        while channel.control.is_set():
            # - start loop latency measurement 
            _SW.start('StrategyContext._process_incoming_data')

            # - waiting for incoming market data
            symbol, d_type, data = channel.queue.get()

            # - process data if handler is registered
            handler = self.__handlers.get(d_type)
            _strategy_trigger_event = handler(self, symbol, data) if handler else None
            if _strategy_trigger_event:
                try:
                    _SW.start('strategy.on_event')
                    self.strategy.on_event(self, _strategy_trigger_event)
                    _fails_counter = 0
                except Exception as strat_error:
                    # - TODO: probably we need some cooldown interval after exception to prevent flooding
                    logger.error(f"[{self.time()}]: Strategy {self.strategy.__class__.__name__} raised an exception: {strat_error}")
                    logger.opt(colors=False).error(traceback.format_exc())

                    #  - we stop execution after let's say maximal number of errors in a row
                    if (_fails_counter := _fails_counter + 1) >= StrategyContext.MAX_NUMBER_OF_STRATEGY_FAILURES:
                        logger.error("STRATEGY FAILURES IN THE ROW EXCEEDED MAX ALLOWED NUMBER - STOPPING ...")
                        channel.stop()
                        break
                finally:
                    _SW.stop('strategy.on_event')

            # - stop loop latency measurement 
            _SW.stop('StrategyContext._process_incoming_data')

            # - notify poition and portfolio loggers
            self._logging.notify(self.time())

        _SW.stop('StrategyContext._process_incoming_data')
        logger.info("(StrategyContext) Market data processing stopped")

    def _processing_hist_bar(self, symbol: str, bar: Bar) -> TriggerEvent | None:
        # - processing single historical bar
        #   here it just updates cache - historical bar can't trigger strategy logic
        self._cache.update_by_bar(symbol, bar)
        return None

    def _processing_hist_bars(self, symbol: str, bars: List[Bar]) -> TriggerEvent | None:
        # - processing historical bars as list
        for b in bars:
            self._processing_hist_bar(symbol, b)
        return None

    def _processing_bar(self, symbol: str, bar: Bar) -> TriggerEvent | None:
        # - processing current bar's update
        self._cache.update_by_bar(symbol, bar)
        if self._trig_on_bar:
            t = self.exchange_service.time().item()
            _time_to_trigger = t % self._trig_bar_freq_nsec >= self._trig_interval_in_bar_nsec
            if _time_to_trigger: 
                # we want to trigger only first one - not every
                if not self._current_bar_trigger_processed:
                    self._current_bar_trigger_processed = True
                    return TriggerEvent(self.time(), 'bar', symbol, bar)
            else:
                self._current_bar_trigger_processed = False
        return None

    def _processing_trade(self, symbol: str, trade: Trade) -> TriggerEvent | None:
        if self._trig_on_trade:
            return TriggerEvent(self.time(), 'trade', symbol, trade)
        return None

    def _processing_quote(self, symbol: str, quote: Quote) -> TriggerEvent | None:
        # - TODO: here we can apply throttlings or filters
        #  - let's say we can skip quotes if bid & ask is not changed
        #  - or we can collect let's say N quotes before sending to strategy
        if self._trig_on_quote:
            return TriggerEvent(self.time(), 'quote', symbol, quote)
        return None

    def _processing_event(self, symbol: str, event: EventFromExchange) -> TriggerEvent | None:
        if self._trig_on_time and event.type == 'time':
            return TriggerEvent(self.time(), 'time', symbol, event)
        return None

    @_SW.watch('StrategyContext')
    def _processing_order(self, symbol: str, order: Order) -> TriggerEvent | None:
        logger.debug(f"[{order.id} / {order.client_id}] : {order.type} {order.side} {order.quantity} of {symbol} { (' @ ' + str(order.price)) if order.price else '' } -> [{order.status}]")
        # - check if we want to trigger any strat's logic on order
        return None

    @_SW.watch('StrategyContext')
    def _processing_deals(self, symbol: str, deals: List[Deal]) -> TriggerEvent | None:
        # - log deals in storage
        self._logging.save_deals(symbol, deals)
        for d in deals:
            logger.debug(f"Executed {d.amount} @ {d.price} of {symbol} for order {d.order_id}")
        return None

    def ohlc(self, instrument: str | Instrument, timeframe: str) -> OHLCV:
        return self._cache.get_ohlcv(instrument if isinstance(instrument, str) else instrument.symbol, timeframe)

    def _create_and_update_positions(self):
        for instrument in self.instruments:
            symb = instrument.symbol
            self.positions[symb] = self.exchange_service.get_position(instrument)
            # - check if we need any aux instrument for calculating pnl ?
            aux = lookup.find_aux_instrument_for(instrument, self.exchange_service.get_base_currency())
            if aux is not None:
                instrument._aux_instrument = aux
                self.instruments.append(aux)
                self.positions[aux.symbol] = self.exchange_service.get_position(aux)

    def start(self):
        if self._is_initilized :
            raise ValueError("Strategy is already started !")

        # - create positions for instruments
        self._create_and_update_positions()

        # - get actual positions from exchange
        _symbols = []
        for instr in self.instruments:
            # process instruments - need to find convertors etc
            self._cache.init_ohlcv(instr.symbol)
            _symbols.append(instr.symbol)

        # - create incoming market data processing
        # self._thread_data_loop = AsyncioThreadRunner(self.data_provider.get_communication_channel())
        self._thread_data_loop = Thread(
            target=self._process_incoming_data, 
            args=(self.data_provider.get_communication_channel(),), 
            daemon=True
        )
        # self._thread_data_loop.add(self._process_incoming_market_data)

        # - subscribe to market data
        logger.info(f"(StrategyContext) Subscribing to {self._market_data_subcription_type} updates using {self._market_data_subcription_params} for \n\t{_symbols} ")
        self.data_provider.subscribe(self._market_data_subcription_type, _symbols, **self._market_data_subcription_params)

        # - initialize strategy loggers
        self._logging.initialize(
            self.time(), self.positions, self.exchange_service.get_account().get_balances()
        )

        # - initialize strategy (should we do that after any first market data received ?)
        if not self._is_initilized:
            try:
                self.strategy.on_start(self)
                self._is_initilized = True
            except Exception as strat_error:
                logger.error(f"(StrategyContext) Strategy {self.strategy.__class__.__name__} raised an exception in on_start: {strat_error}")
                logger.error(traceback.format_exc())
                return

        self._thread_data_loop.start()
        logger.info("(StrategyContext) strategy is started")

    def stop(self):
        if self._thread_data_loop:
            self.data_provider.get_communication_channel().stop()
            self._thread_data_loop.join()
            try:
                self.strategy.on_stop(self)
            except Exception as strat_error:
                logger.error(f"(StrategyContext) Strategy {self.strategy.__class__.__name__} raised an exception in on_stop: {strat_error}")
                logger.error(traceback.format_exc())
            self._thread_data_loop = None

        # - close logging
        self._logging.close()
        for l in _SW.latencies.keys():
            logger.info(f"\t<w>{l}</w> took <r>{_SW.latency_sec(l):.7f}</r> secs")

    def populate_parameters_to_strategy(self, strategy: IStrategy, **kwargs):
        _log_info = ""
        for k,v in kwargs.items():
            if k.startswith('_'):
                raise ValueError("Internal variable can't be set from external parameter !")
            if hasattr(strategy, k):
                strategy.__dict__[k] = v
                _log_info += f"\n\tset <green>{k}</green> <- <red>{v}</red>"
        if _log_info:
            logger.info("(StrategyContext) set strategy parameters:" + _log_info)

    def time(self) -> dt_64:
        return self.exchange_service.time()

    def _generate_order_client_id(self, symbol: str) -> str:
        self.__order_id += 1
        return self.__strategy_id + symbol + '_' + str(self.__order_id)

    @_SW.watch('StrategyContext')
    def trade(self, instr_or_symbol: Instrument | str, amount:float, price: float|None=None, time_in_force='gtc') -> Order:
        _SW.start('send_order')
        instrument: Instrument | None = self._symb_to_instr.get(instr_or_symbol) if isinstance(instr_or_symbol, str) else instr_or_symbol
        if instrument is None:
            raise ValueError(f"Can't find instrument for symbol {instr_or_symbol}")

        # - adjust size
        size_adj = _round_down_at_min_qty(abs(amount), instrument.min_size)
        if size_adj == 0:
            raise ValueError(f"Attempt to trade size {abs(amount)} less than minimal allowed {instrument.min_size} !")

        side = 'buy' if amount > 0 else 'sell'
        type = 'limit' if price is not None else 'market'
        logger.info(f"(StrategyContext) sending {type} {side} for {size_adj} of {instrument.symbol} ...")
        client_id = self._generate_order_client_id(instrument.symbol)
        order = self.exchange_service.send_order(instrument, side, type, size_adj, price, time_in_force=time_in_force, client_id=client_id) 

        return order

    @_SW.watch('StrategyContext')
    def cancel(self, instr_or_symbol: Instrument | str):
        instrument: Instrument | None = self._symb_to_instr.get(instr_or_symbol) if isinstance(instr_or_symbol, str) else instr_or_symbol
        if instrument is None:
            raise ValueError(f"Can't find instrument for symbol {instr_or_symbol}")
        for o in self.exchange_service.get_orders(instrument.symbol):
            self.exchange_service.cancel_order(o.id)

    def quote(self, symbol: str) -> Quote | None:
        return self.data_provider.get_quote(symbol)

    def get_capital(self) -> float:
        return self.exchange_service.get_capital()

    def get_reserved(self, instrument: Instrument) -> float:
        return self.exchange_service.get_account().get_reserved(instrument)


import ccxt.pro as cxp
from ccxt.async_support.base.ws.client import Client
from ccxt.async_support.base.ws.cache import ArrayCacheByTimestamp


class BinanceQV(cxp.binance):
    """
    Extended binance exchange to provide quote asset volumes support
    """
    def parse_ohlcv(self, ohlcv, market=None):
        """
        [
            1499040000000,      // Kline open time                   0
            "0.01634790",       // Open price                        1
            "0.80000000",       // High price                        2
            "0.01575800",       // Low price                         3
            "0.01577100",       // Close price                       4
            "148976.11427815",  // Volume                            5
            1499644799999,      // Kline Close time                  6
            "2434.19055334",    // Quote asset volume                7
            308,                // Number of trades                  8
            "1756.87402397",    // Taker buy base asset volume       9
            "28.46694368",      // Taker buy quote asset volume     10
            "0"                 // Unused field, ignore.
        ]
        """
        return [
            self.safe_integer(ohlcv, 0),
            self.safe_number(ohlcv, 1),
            self.safe_number(ohlcv, 2),
            self.safe_number(ohlcv, 3),
            self.safe_number(ohlcv, 4),
            self.safe_number(ohlcv, 5),
            self.safe_number(ohlcv, 7),   # Quote asset volume
            self.safe_number(ohlcv, 10),  # Taker buy quote asset volume
        ]

    def handle_ohlcv(self, client: Client, message):
        event = self.safe_string(message, 'e')
        eventMap = {
            'indexPrice_kline': 'indexPriceKline',
            'markPrice_kline': 'markPriceKline',
        }
        event = self.safe_string(eventMap, event, event)
        kline = self.safe_value(message, 'k')
        marketId = self.safe_string_2(kline, 's', 'ps')
        if event == 'indexPriceKline':
            # indexPriceKline doesn't have the _PERP suffix
            marketId = self.safe_string(message, 'ps')
        lowercaseMarketId = marketId.lower()
        interval = self.safe_string(kline, 'i')
        # use a reverse lookup in a static map instead
        timeframe = self.find_timeframe(interval)
        messageHash = lowercaseMarketId + '@' + event + '_' + interval
        parsed = [
            self.safe_integer(kline, 't'),
            self.safe_float(kline, 'o'),
            self.safe_float(kline, 'h'),
            self.safe_float(kline, 'l'),
            self.safe_float(kline, 'c'),
            self.safe_float(kline, 'v'),
            # - additional fields
            self.safe_float(kline, 'q'), # - quote asset volume
            self.safe_float(kline, 'Q'), # - taker buy quote asset volume
        ]
        isSpot = ((client.url.find('/stream') > -1) or (client.url.find('/testnet.binance') > -1))
        marketType = 'spot' if (isSpot) else 'contract'
        symbol = self.safe_symbol(marketId, None, None, marketType)
        self.ohlcvs[symbol] = self.safe_value(self.ohlcvs, symbol, {})
        stored = self.safe_value(self.ohlcvs[symbol], timeframe)
        if stored is None:
            limit = self.safe_integer(self.options, 'OHLCVLimit', 2)
            stored = ArrayCacheByTimestamp(limit)
            # self.ohlcvs[symbol][timeframe] = stored
        stored.append(parsed)
        client.resolve(stored, messageHash)


"""
   Misc graphics handy utilitites to be used in interactive analysis
"""
import numpy as np
import pandas as pd
import statsmodels.tsa.stattools as st
import matplotlib
import matplotlib.pyplot as plt
from cycler import cycler


DARK_MATLPLOT_THEME = [
    ('backend', 'module://matplotlib_inline.backend_inline'),
    ('interactive', True),
    ('lines.color', '#5050f0'),
    ('text.color', '#d0d0d0'),
    ('axes.facecolor', '#000000'),
    ('axes.edgecolor', '#404040'),
    ('axes.grid', True),
    ('axes.labelsize', 'large'),
    ('axes.labelcolor', 'green'),
    ('axes.prop_cycle', cycler('color', ['#08F7FE', '#00ff41', '#FE53BB', '#F5D300', '#449AcD', 'g',
                                         '#f62841', 'y', '#088487', '#E24A33', '#f01010'])),
    ('legend.fontsize', 'small'),
    ('legend.fancybox', False),
    ('legend.edgecolor', '#305030'),
    ('legend.shadow', False),
    ('lines.antialiased', True),
    ('lines.linewidth', 0.8),  # reduced line width
    ('patch.linewidth', 0.5),
    ('patch.antialiased', True),
    ('xtick.color', '#909090'),
    ('ytick.color', '#909090'),
    ('xtick.labelsize', 'large'),
    ('ytick.labelsize', 'large'),
    ('grid.color', '#404040'),
    ('grid.linestyle', '--'),
    ('grid.linewidth', 0.5),
    ('grid.alpha', 0.8),
    ('figure.figsize', [12.0, 5.0]),
    ('figure.dpi', 80.0),
    ('figure.facecolor', '#050505'),
    ('figure.edgecolor', (1, 1, 1, 0)),
    ('figure.subplot.bottom', 0.125),
    ('savefig.facecolor', '#000000'),
]

LIGHT_MATPLOT_THEME = [
    ('backend', 'module://matplotlib_inline.backend_inline'),
    ('interactive', True),
    ('lines.color', '#101010'),
    ('text.color', '#303030'),
    ('lines.antialiased', True),
    ('lines.linewidth', 1),
    ('patch.linewidth', 0.5),
    ('patch.facecolor', '#348ABD'),
    ('patch.edgecolor', '#eeeeee'),
    ('patch.antialiased', True),
    ('axes.facecolor', '#fafafa'),
    ('axes.edgecolor', '#d0d0d0'),
    ('axes.linewidth', 1),
    ('axes.titlesize', 'x-large'),
    ('axes.labelsize', 'large'),
    ('axes.labelcolor', '#555555'),
    ('axes.axisbelow', True),
    ('axes.grid', True),
    ('axes.prop_cycle', cycler('color', ['#6792E0', '#27ae60', '#c44e52', '#975CC3', '#ff914d', '#77BEDB',
                                         '#303030', '#4168B7', '#93B851', '#e74c3c', '#bc89e0', '#ff711a',
                                         '#3498db', '#6C7A89'])),
    ('legend.fontsize', 'small'),
    ('legend.fancybox', False),
    ('xtick.color', '#707070'),
    ('ytick.color', '#707070'),
    ('grid.color', '#606060'),
    ('grid.linestyle', '--'),
    ('grid.linewidth', 0.5),
    ('grid.alpha', 0.3),
    ('figure.figsize', [8.0, 5.0]),
    ('figure.dpi', 80.0),
    ('figure.facecolor', '#ffffff'),
    ('figure.edgecolor', '#ffffff'),
    ('figure.subplot.bottom', 0.1)
]


def fig(w=16, h=5, dpi=96, facecolor=None, edgecolor=None, num=None):
    """
    Simple helper for creating figure
    """
    return plt.figure(num=num, figsize=(w, h), dpi=dpi, facecolor=facecolor, edgecolor=edgecolor)


def subplot(shape, loc, rowspan=2, colspan=1):
    """
    Some handy grid splitting for plots. Example for 2x2:
    
    >>> subplot(22, 1); plt.plot([-1,2,-3])
    >>> subplot(22, 2); plt.plot([1,2,3])
    >>> subplot(22, 3); plt.plot([1,2,3])
    >>> subplot(22, 4); plt.plot([3,-2,1])

    same as following

    >>> subplot((2,2), (0,0)); plt.plot([-1,2,-3])
    >>> subplot((2,2), (0,1)); plt.plot([1,2,3])
    >>> subplot((2,2), (1,0)); plt.plot([1,2,3])
    >>> subplot((2,2), (1,1)); plt.plot([3,-2,1])

    :param shape: scalar (like matlab subplot) or tuple
    :param loc: scalar (like matlab subplot) or tuple
    :param rowspan: rows spanned
    :param colspan: columns spanned
    """
    isscalar = lambda x: not isinstance(x, (list, tuple, dict, np.ndarray))

    if isscalar(shape):
        if 0 < shape < 100:
            shape = (max(shape // 10, 1), max(shape % 10, 1))
        else:
            raise ValueError("Wrong scalar value for shape. It should be in range (1...99)")

    if isscalar(loc):
        nm = max(shape[0], 1) * max(shape[1], 1)
        if 0 < loc <= nm:
            x = (loc - 1) // shape[1]
            y = loc - 1 - shape[1] * x
            loc = (x, y)
        else:
            raise ValueError("Wrong scalar value for location. It should be in range (1...%d)" % nm)

    return plt.subplot2grid(shape, loc=loc, rowspan=rowspan, colspan=colspan)


def sbp(shape, loc, r=1, c=1):
    """
    Just shortcut for subplot(...) function

    :param shape: scalar (like matlab subplot) or tuple
    :param loc: scalar (like matlab subplot) or tuple
    :param r: rows spanned
    :param c: columns spanned
    :return:
    """
    return subplot(shape, loc, rowspan=r, colspan=c)


def vline(ax, x, c, lw=1, ls='--'):
    x = pd.to_datetime(x) if isinstance(x, str) else x
    if not isinstance(ax, (list, tuple)):
        ax = [ax]
    for a in ax:
        a.axvline(x, 0, 1, c=c, lw=1, linestyle=ls)


def hline(*zs, mirror=True):
    [plt.axhline(z, ls='--', c='r', lw=0.5) for z in zs]
    if mirror:
        [plt.axhline(-z, ls='--', c='r', lw=0.5) for z in zs]


def ellips(ax, x, y, c='r', r=2.5, lw=2, ls='-'):
    """
    Draw ellips annotation on specified plot at (x,y) point
    """
    from matplotlib.patches import Ellipse
    x = pd.to_datetime(x) if isinstance(x, str) else x
    w, h = (r, r) if np.isscalar(r) else (r[0], r[1])
    ax.add_artist(Ellipse(xy=[x, y], width=w, height=h, angle=0, fill=False, color=c, lw=lw, ls=ls))


def set_mpl_theme(theme: str):
    import plotly.io as pio

    if 'dark' in theme.lower():
        pio.templates.default = "plotly_dark"
        for (k, v) in DARK_MATLPLOT_THEME:
            matplotlib.rcParams[k] = v

    elif 'light' in theme.lower():
        pio.templates.default = "plotly_white"
        for (k, v) in LIGHT_MATPLOT_THEME:
            matplotlib.rcParams[k] = v


from typing import Any, Dict, List, Optional, Tuple

from collections import defaultdict
import stackprinter
import traceback

import ccxt
import ccxt.pro as cxp
from ccxt.base.decimal_to_precision import ROUND_UP
from ccxt.base.exchange import Exchange

import numpy as np
import pandas as pd

from qubx import logger, lookup
from qubx.core.account import AccountProcessor
from qubx.core.basics import Instrument, Position, Order, TransactionCostsCalculator, dt_64, Deal
from qubx.core.strategy import IDataProvider, CtrlChannel, IExchangeServiceProvider
from qubx.core.series import TimeSeries, Bar, Trade, Quote
from qubx.impl.utils import EXCHANGE_ALIASES, ccxt_convert_order_info, ccxt_convert_deal_info, ccxt_extract_deals_from_exec, ccxt_restore_position_from_deals


ORDERS_HISTORY_LOOKBACK_DAYS = 30


class CCXTSyncTradingConnector(IExchangeServiceProvider):
    """
    Synchronous instance of trading API
    """
    sync: Exchange

    _fees_calculator: Optional[TransactionCostsCalculator] = None    # type: ignore
    _positions: Dict[str, Position]

    def __init__(self, 
                 exchange_id: str, 
                 account_id: str,
                 base_currency: str | None, commissions: str|None = None, 
                 reserves: Dict[str, float] | None = None,
                 **exchange_auth):
        if base_currency is None:
            raise ValueError("Base currency is not specified !")

        exchange_id = exchange_id.lower()
        exch = EXCHANGE_ALIASES.get(exchange_id, exchange_id)

        if exch not in ccxt.exchanges:
            raise ValueError(f"Exchange {exchange_id} -> {exch} is not supported by CCXT!")

        # - sync exchange
        self.sync: Exchange = getattr(ccxt, exchange_id.lower())(exchange_auth)
        self.acc = AccountProcessor(account_id, base_currency, reserves)

        logger.info(f"{exch.upper()} loading ...")
        self.sync.load_markets()        
        self._sync_account_info(commissions)
        self._positions = self.acc._positions

        # - show reserves info
        for s, v in self.acc.reserved.items():
            logger.info(f" > {v} of {s} is reserved from trading")

    def _sync_account_info(self, default_commissions: str | None):
        logger.info(f'Loading account data for {self.get_name()}')
        self._balance = self.sync.fetch_balance()
        _info = self._balance.get('info')

        # - check what we have on balance: TODO test on futures account
        for k, vol in self._balance['total'].items(): # type: ignore
            if k.lower() == self.acc.base_currency.lower():
                _free = self._balance['free'][self.acc.base_currency]
                self.acc.update_base_balance(vol, vol - _free)

            if vol != 0.0: # - get all non zero balances 
                self.acc.update_balance(k, vol, vol - self._balance['free'].get(k, 0))

        # - try to get account's commissions calculator or set default one
        if _info:
           _fees = _info.get('commissionRates')
           if _fees:
               self._fees_calculator = TransactionCostsCalculator('account', 100*float(_fees['maker']), 100*float(_fees['taker'])) 

        if self._fees_calculator is None:
            if default_commissions:
                self._fees_calculator = lookup.fees.find(self.get_name().lower(), default_commissions)
            else: 
                raise ValueError("Can't get commissions level from account, but default commissions is not defined !")

    def _get_open_orders_from_exchange(self, symbol: str, days_before: int = 60) -> Dict[str, Order]:
        """
        We need only open orders to restore list of active ones in connector 
        method returns open orders sorted by creation time in ascending order
        """
        t_orders_start_ms = ((self.time() - days_before * pd.Timedelta('1d')).asm8.item() // 1000000)
        orders_data = self.sync.fetch_open_orders(symbol, since=t_orders_start_ms)
        orders: Dict[str, Order] = {}
        for o in orders_data:
            order = ccxt_convert_order_info(symbol, o) 
            orders[order.id] = order
        if orders: 
            logger.info(f"{symbol} - loaded {len(orders)} open orders")
        return dict(sorted(orders.items(), key=lambda x: x[1].time, reverse=False))

    def _get_deals_from_exchange(self, symbol: str, days_before: int = 60) -> List[Deal]:
        """
        Load trades for given symbol
        method returns account's trades sorted by creation time in reversed order (latest - first)
        """
        t_orders_start_ms = ((self.time() - days_before * pd.Timedelta('1d')).asm8.item() // 1000000)
        deals_data = self.sync.fetch_my_trades(symbol, since=t_orders_start_ms)
        deals: List[Deal] = [ccxt_convert_deal_info(o) for o in deals_data] # type: ignore
        if deals:
            return list(sorted(deals, key=lambda x: x.time, reverse=False))
        return list()

    def _sync_position_and_orders(self, position: Position) -> Position:
        asset = position.instrument.base
        symbol = position.instrument.symbol
        total_amnts = self._balance['total']
        vol_from_exch = total_amnts.get(asset, total_amnts.get(symbol, 0))

        # - get orders from exchange
        orders = self._get_open_orders_from_exchange(position.instrument.symbol, ORDERS_HISTORY_LOOKBACK_DAYS)
        self.acc.add_active_orders(orders)

        # - get deals from exchange if position is not zero
        if vol_from_exch != 0:
            deals = self._get_deals_from_exchange(symbol)

            # - actualize position
            position = ccxt_restore_position_from_deals(position, vol_from_exch, deals, reserved_amount=self.acc.get_reserved(position.instrument));

        return position

    def get_position(self, instrument: Instrument) -> Position:
        symbol = instrument.symbol

        if symbol not in self._positions:
            position = Position(instrument, self._fees_calculator)  # type: ignore
            position = self._sync_position_and_orders(position)
            self.acc.attach_positions(position)

        return self._positions[symbol] 

    def update_position_price(self, symbol: str, price: float):
        self.acc.update_position_price(self.time(), symbol, price)

    def send_order(
        self, instrument: Instrument, order_side: str, order_type: str, amount: float, price: float | None = None, 
        client_id: str | None = None, time_in_force: str='gtc'
    ) -> Optional[Order]:
        params={}
        symbol = instrument.symbol

        if order_type == 'limit':
            params['timeInForce'] = time_in_force.upper()
            if price is None:
                raise ValueError('Price must be specified for limit order')

        if client_id:
            params['newClientOrderId'] = client_id

        try:
            r: Dict[str, Any] | None = self.sync.create_order(
                symbol, order_type, order_side, amount, price, # type: ignore
                params=params)
        except ccxt.BadRequest as exc:
            logger.error(f"(CCXTSyncTradingConnector::send_order) BAD REQUEST for {order_side} {amount} {order_type} for {symbol} : {exc}")
            raise exc
        except Exception as err:
            logger.error(f"(CCXTSyncTradingConnector::send_order) {order_side} {amount} {order_type} for {symbol} exception : {err}")
            logger.error(traceback.format_exc())
            raise err

        if r is not None:
            order = ccxt_convert_order_info(symbol, r) 
            logger.info(f"(CCXTSyncTradingConnector) New order {order}")
            return order

        return None

    def cancel_order(self, order_id: str) -> Order | None:
        order = None
        if order_id in self.acc._active_orders:
            order = self.acc._active_orders[order_id]
            try:
                logger.info(f"Canceling order {order_id} ...")
                # r = self._task_s(self.exchange.cancel_order(order_id, symbol=order.symbol))
                r = self.sync.cancel_order(order_id, symbol=order.symbol)
            except Exception as err:
                logger.error(f"(CCXTSyncTradingConnector) canceling [{order}] exception : {err}")
                logger.error(traceback.format_exc())
                raise err
        return order

    def time(self) -> dt_64:
        """
        Returns current time in nanoseconds
        """
        return np.datetime64(self.sync.microseconds() * 1000, 'ns')

    def get_name(self) -> str:
        return self.sync.name  # type: ignore

    def _process_execution_report(self, symbol: str, report: Dict[str, Any]) -> Tuple[Order, List[Deal]]:
        order = ccxt_convert_order_info(symbol, report)
        deals = ccxt_extract_deals_from_exec(report)
        self.acc.process_deals(symbol, deals)
        self.acc.process_order(order)
        return order, deals

    def get_orders(self, symbol: str | None = None) -> List[Order]:
        return self.acc.get_orders(symbol)

    def get_base_currency(self) -> str:
        return self.acc.base_currency